use clap::Parser;
use readmap::types::{Content::*, Section};
use tool_helper::{Error, exit_with_error};
use std::{fs::File, io::{BufWriter, Write}, path::PathBuf};

#[derive(Parser)]
#[clap(about = "Opens and scans a MAP file to print extended information", long_about = None)]
struct CommandLine {
    #[clap(value_parser, help="Input MAP file for scannning")]
    input:          Option<PathBuf>
}

fn run_main(cmd: CommandLine) -> Result<(), Error> {
    fn value_to_hex<T: std::fmt::UpperHex>(value: T) -> String {
        return format!("0x{:X}", value);
    }

    fn option_to_hex<T: std::fmt::UpperHex>(value: Option<T>) -> String {
        if let Some(value) = value {
            return value_to_hex(value);
        }

        else {
            return String::from("<None>");
        }
    }

    let     sections = readmap::scan(tool_helper::open_input(cmd.input)?)?;
    let mut file     = tool_helper::open_output_file(&PathBuf::from("./planschi.txt"))?;

    for section in sections {
        fn print_content(tab_level: usize, file: &mut BufWriter<File>, section: Section) -> Result<(), Error> {
            for content in section.content {
                match content {
                    Fill(fill)       => {
                        writeln!(file, "{:>tab_level$}*fill* @{}, {}", ' ', value_to_hex(fill.adr), value_to_hex(fill.size), tab_level=tab_level*4)?;
                    },
                    Section(section) => {
                        writeln!(file, "{:>tab_level$} {} @{}, {}", ' ', section.name, option_to_hex(section.adr), option_to_hex(section.size), tab_level=tab_level*4)?;
                        print_content(tab_level + 1, file, section)?;
                    },
                    Symbol(symbol)   => {
                        writeln!(file, "{:>tab_level$}{} @{}", ' ', symbol.name, value_to_hex(symbol.adr), tab_level=tab_level*4)?;
                    }
                }
            }

            Ok(())
        }

        writeln!(file, "{}: @{}, {}", section.name, option_to_hex(section.adr), option_to_hex(section.size))?;
        print_content(1, &mut file, section)?;
    }
    Ok(())
}

pub fn main() {
    match CommandLine::try_parse() {
        Ok(cmd_line) => {
            match run_main(cmd_line) {
                Ok(_)      => (),
                Err(error) => exit_with_error(error)
            }
        },
        Err(error) => {
            println!("{}", error)
        }
    }
}